// The following were derived from the Platform SDK files:
// Winerror.h, Winnt.h
// DO NOT MODIFY

#define	SE_RESTORE_NAME				"SeRestorePrivilege"  
#define SE_PRIVILEGE_ENABLED		0x00000002  
#define TOKEN_QUERY             	0x00000008
#define TOKEN_ADJUST_PRIVILEGES 	0x00000020  
#define ERROR_PRIVILEGE_NOT_HELD	1314

#ifndef ERROR_SHARING_VIOLATION  
#define ERROR_SHARING_VIOLATION		32
#endif

// The following constants are custom defines, these may be modified
#define	ERROR_TOKEN_NOT_ADJUSTED	220
#define	ERROR_LOOKUP_PRIVILEGE		221
#define ERROR_LOADING_ADVAPI32		230
#define ERROR_UNLOADING_ADVAPI32	231
#define ERROR_HIVE_FILE_NOT_FOUND	232
#define ERROR_OPEN_PROCESS_TOKEN	233
#define ERROR_INVALID_PROFILE		234
#define ERROR_TOKEN_NOT_DISABLED	235
#define	ERROR_INVALID_REGDB			236
#define	REGDB_SUBKEY				1
#define	REGDB_VALUE					2

#define	DEFAULT_USER_PROFILE	1
#define OTHER_USER_PROFILE		2
#define ALL_USERS_PROFILE		3

/////////////////////// Typedef structures ///////////////////////

// LUID Data Type 
// http://msdn.microsoft.com/library/en-us/security/Security/luid.asp
typedef LUID
begin
	NUMBER 	LowPart;		// Low order bits
	LONG	HighPart;		// High order bits
end;

// LUID_AND_ATTRIBUTES Data Type
// http://msdn.microsoft.com/library/en-us/security/Security/luid_and_attributes.asp
typedef LUID_AND_ATTRIBUTES
begin
	LUID 	Luid;      		// Specifies the Locally Unique Identifier (LUID)
	LONG	Attributes;		// Specifies the attributes (See Platform SDK for values)
end;

// TOKEN_PRIVILEGES Data Type
// http://msdn.microsoft.com/library/en-us/security/Security/token_privileges.asp
typedef TOKEN_PRIVILEGES
begin
	LONG					PrivilegeCount;	// Number of items in the Privileges array
	LUID_AND_ATTRIBUTES  	Privileges;		// Array of LUID_AND_ATTRIBUTES
	// NOTE:  InstallScript typedef structs do not support arrays (LISTS are _not_ arrays).
	//        For the purpose of this header file and calling functions, no more than one
	//        element is required.  If you wish to use more, please restructure this struct.
end;                                                                                  

//////////////////// DLL Function Prototypes /////////////////////

// GetCurrentProcess: Get the current process's pseudo handle
// The returned token is required by OpenProcessToken
prototype NUMBER kernel32.GetCurrentProcess();                   

// CloseHandle: Close an open handle; no need to UseDll the kernel32 library
// http://msdn.microsoft.com/library/en-us/helplib/handobj_289x.asp
prototype BOOL kernel32.CloseHandle (
	NUMBER			// nHandle	[in] Handle
);

// LookupPrivilegeValue (ANSI version): Retrieve LUID used to represent the required privilege
// http://msdn.microsoft.com/library/en-us/security/Security/lookupprivilegevalue.asp
// On IS6, no need to UseDll advapi32 library
prototype BOOL advapi32.LookupPrivilegeValueA (
	BYVAL STRING,	// SystemName 	[in] system name; "" = local system
	BYVAL STRING,	// Name 		[in] privilege constant string (See Winnt.h)
	LUID POINTER	// pLuid 		[out] Pointer to the Luid of the specified system
					// Try without the "LUID"
); 

// OpenProcessToken: Opens the access token for a specified token and specify Access Rights
// http://msdn.microsoft.com/library/en-us/security/Security/openprocesstoken.asp 
prototype BOOL advapi32.OpenProcessToken (
	NUMBER,			// ProcessHandle [in]  Return value from GetCurrentProcess()
	LONG,			// DesiredAccess [in]
	POINTER			// TokenHandle	 [out] Pointer to the returned handle ID (used by AdjustTokenPrivileges)
);

// AdjustTokenPrivileges: Enables/Disables priviliges in the token specified by TokenHandle
// http://msdn.microsoft.com/library/en-us/security/Security/adjusttokenprivileges.asp
prototype BOOL advapi32.AdjustTokenPrivileges (
	NUMBER,							// TokenHandle	 [in] Handle to access token from OpenProcessToken
	BOOL,							// DisablePriv	 [in] Disable all privileges
	TOKEN_PRIVILEGES POINTER,		// pTokenPriv	 [in] Pointer to TOKEN_PRIVILEGES struct
	LONG,							// BufferLen	 [in] 0
	TOKEN_PRIVILEGES POINTER,		// pPreviousPriv [out] Pointer to a TOKEN_PRIVILEGES struct
	POINTER							// ReturnLen	 [out] Pointer to pPreviousPriv buffer
);

// RegLoadKey (ANSI version): Creates a subkey under the ParentHive and stores a registry hive file   
// NOTE: Process must have SE_RESTORE_NAME privilege to perform the function call
// http://msdn.microsoft.com/library/en-us/sysinfo/regapi_19h5.asp
prototype LONG advapi32.RegLoadKeyA (
	NUMBER,			// ParentHive	[in] HKEY_LOCAL_MACHINE or HKEY_USERS
	BYVAL STRING,	// sSubKey		[in] Subkey name
	BYVAL STRING	// sFile		[in] NTUSER.DAT path
); 

// RegUnloadKey (ANSI version): Unloads the specified subkey under the ParentHive
// NOTE: Process must have SE_RESTORE_NAME privilege to perform the function call 
// http://msdn.microsoft.com/library/en-us/sysinfo/regapi_81kp.asp
prototype LONG advapi32.RegUnLoadKeyA (
	NUMBER,			// ParentHive	[in] HKEY_LOCAL_MACHINE or HKEY_USERS
	BYVAL STRING	// sSubKey		[in] Subkey name to unload
);

////////////////// Custom Function Prototypes ///////////////////
prototype NUMBER __NTProfResolveProfileDir ( BYREF STRING );	
prototype NUMBER __NTProfRegLoad ( NUMBER , STRING , STRING ); 
prototype NUMBER __NTProfRegUnload ( NUMBER , STRING );
prototype NUMBER _NTProfCopyFiles ( STRING , STRING , NUMBER , NUMBER );
prototype NUMBER _NTProfDeleteFiles ( STRING , NUMBER );
prototype NUMBER _NTProfDeleteDir ( STRING , NUMBER , NUMBER );
prototype NUMBER _NTProfRegHiveDelete ( NUMBER , STRING , STRING , NUMBER ); 
prototype NUMBER _NTProfRegHiveSetKeyValue ( STRING , STRING , NUMBER , STRING , NUMBER , NUMBER , BOOL ); 

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   _NTProfCopyFiles
//
//  PURPOSE:	Copies files to either the DefaultUser or all other current
//				local user profiles
//
//	PARAMETERS:	szSourceFiles	- Path to the source files to be copied to the profiles' folders
//								  You can use wild cards
//				szDestinationFolder - Path to the destination folder (relative to the profile's folder)
//				nzXcopyOperator	- See "nOp" parameter for the InstallScript XcopyFile function
//				nzProfileSwitch - Numeric switch to indicate which profile to copy the file(s) to
//								  Possible values are DEFAULT_USER_PROFILE or OTHER_USER_PROFILE 
//								  or ALL_USERS_PROFILE
//
//	RETURNS:	 0	- The function was successful
//				-1	- Non-NT machine
//				-2  - Invalid szProfileSwitch value                                                 
//
//	NOTE:		- Do not combine szProfileSwitches with BitOR operators (|)
//				- You can use BitOR operators in nzXcopyOperator
//				- Do not use this function to copy shortcuts
//
//////////////////////////////////////////////////////////////////////////////
function NUMBER _NTProfCopyFiles ( szSourceFiles , szDestinationFolder , nzXCopyOperator , nzProfileSwitch )   
	STRING sProfileListKey, sUserProfileDir, sRegProfileValue, sTempUserKeyValue; 
	STRING sTempProfileSID, sTempUserProfileDir; 
	LIST   lUserProfileSIDs;
	NUMBER nType, nSize, nResult;
begin
	if !( SYSINFO.WINNT.bWinNT ) then
		return -1;
	endif;
	
	RegDBSetDefaultRoot ( HKEY_LOCAL_MACHINE );
	sProfileListKey = "Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList";
	
	if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) || ( nzProfileSwitch = ALL_USERS_PROFILE ) then
		if ( SYSINFO.WINNT.bWinNT4 ) then
			if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
				// Get Default User Profile (all changes to this profile will be inherited to 
				// newly created profiles on the target machine)
				sUserProfileDir = WINDIR ^ "Profiles\\Default User";
			else
				// Get the All Users Profile
				sUserProfileDir = WINDIR ^ "Profiles\\All Users";
			endif;
		else 
			// Get the root folder containing all profiles
			RegDBGetKeyValueEx ( sProfileListKey , "ProfilesDirectory" , nType , sRegProfileValue , nSize );
	        __NTProfResolveProfileDir ( sRegProfileValue );       
			if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
				// Get Default User profile folder
				RegDBGetKeyValueEx (sProfileListKey, "DefaultUserProfile", nType, sTempUserKeyValue, nSize);
			else
				// Get the All Users Profile
				RegDBGetKeyValueEx (sProfileListKey, "AllUsersProfile", nType, sTempUserKeyValue, nSize);
			endif;
			// Append sRegProfileValue^sDefaultUserProfileDir 
			sUserProfileDir = sRegProfileValue ^ sTempUserKeyValue;
		endif;
		
		// Copy files to the Default User folder 
		XCopyFile ( szSourceFiles , sUserProfileDir ^ szDestinationFolder , nzXCopyOperator );
		RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
		return 0;
	endif;
	
	// Get other User Profiles and copy the file to them  
	if ( nzProfileSwitch = OTHER_USER_PROFILE ) then
		// Enumerate User Profiles and copy the files 
		lUserProfileSIDs = ListCreate ( STRINGLIST );
		RegDBQueryKey ( sProfileListKey , REGDB_KEYS , lUserProfileSIDs );
		nResult = ListGetFirstString ( lUserProfileSIDs , sTempProfileSID );
		while ( nResult != END_OF_LIST )
			RegDBGetKeyValueEx (sProfileListKey ^ sTempProfileSID, "ProfileImagePath", nType , sTempUserProfileDir , nSize);
			__NTProfResolveProfileDir (sTempUserProfileDir);                     
			XCopyFile ( szSourceFiles , sTempUserProfileDir ^ szDestinationFolder , nzXCopyOperator );
			nResult = ListGetNextString ( lUserProfileSIDs , sTempProfileSID );
		endwhile;      
		
		RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
		return 0;
	endif;
	
	// Invalid szProfileSwitch value
	return -2;
end;

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   _NTProfDeleteFiles
//
//  PURPOSE:	Delete specific files within a specified profile switch. May be used
//				in scripted uninstallations where files that were inherited from the Default
//				User profile are not logged for uninstallation
//
//	USED IN:	Scripted uninstallations
//
//	PARAMETERS:	szFiles			- Path to the files to be deleted (relative to the profile's folder)
//				nzProfileSwitch - Numeric switch to indicate which profile to delete the file(s) from
//								  Possible values are DEFAULT_USER_PROFILE or OTHER_USER_PROFILE
//								  or ALL_USERS_PROFILE
//
//	RETURNS:	 0	- The function was successful
//				-1	- Non-NT machine
//				-2  - Invalid nzProfileSwitch value
//
////////////////////////////////////////////////////////////////////////////// 
function NUMBER _NTProfDeleteFiles ( szFiles , nzProfileSwitch ) 
	STRING sProfileListKey, sUserProfileDir, sRegProfileValue, sTempUserKeyValue; 
	STRING sTempProfileSID, sTempUserProfileDir; 
	LIST   lUserProfileSIDs;
	NUMBER nType, nSize, nResult, nFindFileResult;
	STRING svReturnPath, svReturnWildCard ,svFullFileName;
begin
	if !( SYSINFO.WINNT.bWinNT ) then
		return -1;
	endif; 
	
	RegDBSetDefaultRoot ( HKEY_LOCAL_MACHINE );
	sProfileListKey = "Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList";
	
	if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) || ( nzProfileSwitch = ALL_USERS_PROFILE ) then
		if ( SYSINFO.WINNT.bWinNT4 ) then
			if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
				// Get Default User Profile (all changes to this profile will be inherited to 
				// newly created profiles on the target machine)
				sUserProfileDir = WINDIR ^ "Profiles\\Default User";
			else
				// Get the All Users Profile
				sUserProfileDir = WINDIR ^ "Profiles\\All Users";
			endif;
		else 
			// Get the root folder containing all profiles
			RegDBGetKeyValueEx ( sProfileListKey , "ProfilesDirectory" , nType , sRegProfileValue , nSize );
	        __NTProfResolveProfileDir ( sRegProfileValue );       
			if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
				// Get Default User profile folder
				RegDBGetKeyValueEx (sProfileListKey, "DefaultUserProfile", nType, sTempUserKeyValue, nSize);
			else
				// Get the All Users Profile
				RegDBGetKeyValueEx (sProfileListKey, "AllUsersProfile", nType, sTempUserKeyValue, nSize);
			endif;
			// Append sRegProfileValue^sDefaultUserProfileDir 
			sUserProfileDir = sRegProfileValue ^ sTempUserKeyValue;
		endif;
		
		if ( szFiles % "*" ) then
			// Separate the path from the file
			ParsePath ( svReturnPath , sUserProfileDir ^ szFiles , PATH );
			ParsePath ( svReturnWildCard , sUserProfileDir ^ szFiles , FILENAME );
			// Reset the matching file's attributes to avoid errors on DeleteFile
			nFindFileResult = FindAllFiles( svReturnPath , svReturnWildCard , svFullFileName , RESET );
			while ( nFindFileResult = 0 )		// (<0) indicates no more matching files
				SetFileInfo ( svFullFileName , FILE_ATTRIBUTE , FILE_ATTR_NORMAL , "" );
				nFindFileResult = FindAllFiles ( svReturnPath , svReturnWildCard , svFullFileName , CONTINUE );
			endwhile;
		endif;	
		// Perform the DeleteFileFunction for every file that matches
		DeleteFile ( sUserProfileDir ^ szFiles );

 		RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
 		return 0;
	endif;
	
	// Get other User Profiles and copy the file to them  
	if ( nzProfileSwitch = OTHER_USER_PROFILE ) then
		// Enumerate User Profiles and copy the files 
		lUserProfileSIDs = ListCreate ( STRINGLIST );
		RegDBQueryKey ( sProfileListKey , REGDB_KEYS , lUserProfileSIDs );
		nResult = ListGetFirstString ( lUserProfileSIDs , sTempProfileSID );
		while ( nResult != END_OF_LIST )
			RegDBGetKeyValueEx (sProfileListKey ^ sTempProfileSID, "ProfileImagePath", nType , sTempUserProfileDir , nSize);
			__NTProfResolveProfileDir (sTempUserProfileDir);                     
			if ( szFiles % "*" ) then
				// Separate the path from the file
				ParsePath ( svReturnPath , sTempUserProfileDir ^ szFiles , PATH );
				ParsePath ( svReturnWildCard , sTempUserProfileDir ^ szFiles , FILENAME );
				// Reset the matching file's attributes to avoid errors on DeleteFile
				nFindFileResult = FindAllFiles( svReturnPath , svReturnWildCard , svFullFileName , RESET );
				while ( nFindFileResult = 0 )		// (<0) indicates no more matching files
					SetFileInfo ( svFullFileName , FILE_ATTRIBUTE , FILE_ATTR_NORMAL , "" );
					nFindFileResult = FindAllFiles ( svReturnPath , svReturnWildCard , svFullFileName , CONTINUE );
				endwhile;
			endif;	
			// Perform the DeleteFileFunction for every file that matches
			DeleteFile ( sTempUserProfileDir ^ szFiles );
   			nResult = ListGetNextString ( lUserProfileSIDs , sTempProfileSID );
		endwhile;      
		
		RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
		return 0;
	endif;
	
	// Invalid szProfileSwitch value
	return -2;

//	SetFileInfo ( sUserProfileDir ^ szFiles , FILE_ATTRIBUTE , FILE_ATTR_NORMAL , "" );
//	DeleteFile ( sUserProfileDir ^ szFiles );
end;

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   _NTProfDeleteDir
//
//  PURPOSE:	Delete specific directory within a specified profile switch. May be used
//				in scripted uninstallations where files that were inherited from the Default
//				User profile are not logged for uninstallation
//
//	USED IN:	Scripted uninstallations
//
//	PARAMETERS:	szDir			- Name of the folder to be deleted (relative to the profile's folder)
//				nFlag			- Special deletion options (see DeleteDir InstallScript function)
//				nzProfileSwitch - Numeric switch to indicate which profile to delete the file(s) from
//								  Possible values are DEFAULT_USER_PROFILE or OTHER_USER_PROFILE
//								  or ALL_USERS_PROFILE
//
//	RETURNS:	 0	- The function was successful
//				-1	- Non-NT machine
//				-2  - Invalid nzProfileSwitch value
//
////////////////////////////////////////////////////////////////////////////// 
function NUMBER _NTProfDeleteDir ( szDir , nFlag , nzProfileSwitch ) 
	STRING sProfileListKey, sUserProfileDir, sRegProfileValue, sTempUserKeyValue; 
	STRING sTempProfileSID, sTempUserProfileDir; 
	LIST   lUserProfileSIDs;
	NUMBER nType, nSize, nResult;
begin
	if !( SYSINFO.WINNT.bWinNT ) then
		return -1;
	endif; 
	
	RegDBSetDefaultRoot ( HKEY_LOCAL_MACHINE );
	sProfileListKey = "Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList";
	
	if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) || ( nzProfileSwitch = ALL_USERS_PROFILE ) then
		if ( SYSINFO.WINNT.bWinNT4 ) then
			if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
				// Get Default User Profile (all changes to this profile will be inherited to 
				// newly created profiles on the target machine)
				sUserProfileDir = WINDIR ^ "Profiles\\Default User";
			else
				// Get the All Users Profile
				sUserProfileDir = WINDIR ^ "Profiles\\All Users";
			endif;
		else 
			// Get the root folder containing all profiles
			RegDBGetKeyValueEx ( sProfileListKey , "ProfilesDirectory" , nType , sRegProfileValue , nSize );
	        __NTProfResolveProfileDir ( sRegProfileValue );       
			if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
				// Get Default User profile folder
				RegDBGetKeyValueEx (sProfileListKey, "DefaultUserProfile", nType, sTempUserKeyValue, nSize);
			else
				// Get the All Users Profile
				RegDBGetKeyValueEx (sProfileListKey, "AllUsersProfile", nType, sTempUserKeyValue, nSize);
			endif;
			// Append sRegProfileValue^sDefaultUserProfileDir 
			sUserProfileDir = sRegProfileValue ^ sTempUserKeyValue;
		endif;
		
		// Perform the DeleteDir function
		DeleteDir ( sUserProfileDir ^ szDir , nFlag );
		
 		RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
 		return 0;
	endif;
	
	// Get other User Profiles and copy the file to them  
	if ( nzProfileSwitch = OTHER_USER_PROFILE ) then
		// Enumerate User Profiles and copy the files 
		lUserProfileSIDs = ListCreate ( STRINGLIST );
		RegDBQueryKey ( sProfileListKey , REGDB_KEYS , lUserProfileSIDs );
		nResult = ListGetFirstString ( lUserProfileSIDs , sTempProfileSID );
		while ( nResult != END_OF_LIST )
			RegDBGetKeyValueEx (sProfileListKey ^ sTempProfileSID, "ProfileImagePath", nType , sTempUserProfileDir , nSize);
			__NTProfResolveProfileDir (sTempUserProfileDir);                     
			
			// Perform the DeleteFileFunction for every file that matches
			DeleteDir ( sTempUserProfileDir ^ szDir , nFlag );
   			nResult = ListGetNextString ( lUserProfileSIDs , sTempProfileSID );
		endwhile;      
		
		RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
		return 0;
	endif;
	
	// Invalid szProfileSwitch value
	return -2;

end;

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   __NTProfResolveProfileDir
//
//  PURPOSE:	Resolve a ProfileImagePathString to its absolute value
//
//	USED IN:	_NTProfCopyFiles
//
//	PARAMETERS:	szProfileImagePath 	- Prifile path to resolve
//
//	RETURNS:	0					- The function was successful
//				<0					- Generic failure
//
//////////////////////////////////////////////////////////////////////////////
function NUMBER __NTProfResolveProfileDir (szProfileImagePath)
	STRING	sTempKeyValue, sEnvVarName, sEnvVarValue;                  
	NUMBER	nResult;
	LIST	lTempKeyValue;
begin
	lTempKeyValue = ListCreate(STRINGLIST);
	StrGetTokens (lTempKeyValue , szProfileImagePath , "\\");
 	nResult = ListGetFirstString (lTempKeyValue , sTempKeyValue);              
	while ( nResult != END_OF_LIST )
		if ( sTempKeyValue % "%" ) then	// It is an environment variable
			StrSub ( sEnvVarName , sTempKeyValue , 1 , (StrLengthChars(sTempKeyValue) - 2)); // Get env name
			GetEnvVar ( sEnvVarName , sEnvVarValue );
			ListSetCurrentString (lTempKeyValue, sEnvVarValue);	// replace envVar with envValue
		endif;
		nResult = ListGetNextString ( lTempKeyValue , sTempKeyValue);   
	endwhile;                                            

	// Re-assemble the list	                                          
	szProfileImagePath = "";
	StrPutTokens (lTempKeyValue , szProfileImagePath , "\\" , FALSE); 
	ListDestroy (lTempKeyValue); 
	return 0;
end;                                                                             

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   __NTProfRegLoad
//
//  PURPOSE:	Creates a subkey under HKEY_LOCAL_MACHINE or HKEY_USERS and loads a hive file
//				under the created subkey.
//
//	USED IN:	_NTProfRegHiveDeleteKey
//
//	PARAMETERS:	nKey		- Root key to use (HKEY_LOCAL_MACHINE or HKEY_USERS only)	
//				sRegPath	- Subkey name to load the hive under
//				sFile		- Fully qualified path to the hive file (NTUSER.DAT) to load
//
//	RETURNS:	0							- Success
//				ERROR_HIVE_FILE_NOT_FOUND 	- Hive File (NTUSER.DAT not found)
//				ERROR_LOADING_ADVAPI32		- Unable to load ADVAPI32.DLL
//				ERROR_OPEN_PROCESS_TOKEN	- Unable to run OpenProcessToken API function
//				ERROR_LOOKUP_PRIVILEGE		- Unable to run LookupPrivilegeValueA API function
//				ERROR_TOKEN_NOT_ADJUSTED	- Unable to run AdjustTokenPrivileges API function
//				ERROR_SHARING_VIOLATION		- Hive file is already loaded and locked
//				ERROR_PRIVILEGE_NOT_HELD	- AdjustTokenPrivileges passed, but gave the wrong privilege
//
////////////////////////////////////////////////////////////////////////////// 
function NUMBER __NTProfRegLoad ( nKey , sRegPath , sFile )
	BOOL				nRetVal;
	LONG				lRegRetVal, lBuffer;
	POINTER				hToken;
	TOKEN_PRIVILEGES	tp; 
	LUID_AND_ATTRIBUTES lna;
	LUID				localLuid;
begin     
	// Check for hive file existance
	if (!Is(FILE_EXISTS , sFile)) then
		return ERROR_HIVE_FILE_NOT_FOUND;
	endif;                          
	
	// Get the process token
	nRetVal = OpenProcessToken ( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken );
	if (!nRetVal) then
		// SprintfBox (MB_OK, "OpenProcess Token", "Fail" );    
		return ERROR_OPEN_PROCESS_TOKEN;
	endif;                                                    
	
	// Get the Luid
	nRetVal = LookupPrivilegeValueA ( "", "SeRestorePrivilege",	&localLuid );
	if (!nRetVal) then
		//SprintfBox (MB_OK , "LookupPrivilegeValueA FAIL" , "localLuid = %d", localLuid );
		return ERROR_LOOKUP_PRIVILEGE;
	endif; 
	
	// Populate the TOKEN_PRIVILEGES struct, tp
	tp.PrivilegeCount = 1;            
	tp.Privileges.Luid.LowPart = localLuid.LowPart; 
	tp.Privileges.Luid.HighPart = localLuid.HighPart;
	tp.Privileges.Attributes = SE_PRIVILEGE_ENABLED; 
	
	// Apply TOKEN PRIVILEGES
	nRetVal = AdjustTokenPrivileges ( hToken , FALSE , &tp , 0 , NULL , NULL );
	if (!nRetVal) then
		//SprintfBox (MB_OK, "AdjustTokenPrivileges FAIL","");
		return ERROR_TOKEN_NOT_ADJUSTED;
	endif;

	// Close Handle
	CloseHandle ( hToken );
	
	// Load the Hive
    lRegRetVal = RegLoadKeyA ( nKey , sRegPath , sFile );
    if ( lRegRetVal != 0 ) then 
    	// Two known errors have been identified to date:
    	//	- ERROR_SHARING_VIOLATION: The .DAT file has already been loaded to a hive
    	//  - ERROR_PRIVILEGE_NOT_HELD: AdjustTokenPrivileges passed, but gave the wrong privilege
    	return lRegRetVal;
    endif;
    
	return 0;
end;

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   __NTProfRegUnload
//
//  PURPOSE:	Unloads a subkey under HKEY_LOCAL_MACHINE or HKEY_USERS 
//
//	USED IN:	_NTProfRegHiveDeleteKey
//
//	PARAMETERS:	nKey		- Root key to use (HKEY_LOCAL_MACHINE or HKEY_USERS only)	
//				sRegPath	- Subkey name to load the hive under
//
//	RETURNS:	0							- Success
//				ERROR_LOADING_ADVAPI32		- Unable to load ADVAPI32.DLL                               
//				ERROR_OPEN_PROCESS_TOKEN	- Unable to run OpenProcessToken API function               
//				ERROR_LOOKUP_PRIVILEGE		- Unable to run LookupPrivilegeValueA API function          
//				ERROR_TOKEN_NOT_ADJUSTED	- Unable to run AdjustTokenPrivileges API function          
//				ERROR_SHARING_VIOLATION		- Hive file is already loaded and locked                    
//				ERROR_PRIVILEGE_NOT_HELD	- AdjustTokenPrivileges passed, but gave the wrong privilege
//
////////////////////////////////////////////////////////////////////////////// 
function NUMBER __NTProfRegUnload ( nKey , sRegPath )
	BOOL				nRetVal;
	LONG				lRegRetVal, lBuffer;
	POINTER				hToken;
	TOKEN_PRIVILEGES	tp, tp2; 
	LUID_AND_ATTRIBUTES lna;
	LUID				localLuid;
begin     

	// Unload the specified hive  
	lRegRetVal = RegUnLoadKeyA ( nKey , sRegPath );
    if ( lRegRetVal != 0 ) then 
    	return lRegRetVal;
    endif;

	// REMOVE PRIVILEGE
	// Get the process token
	nRetVal = OpenProcessToken ( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken );
	if (!nRetVal) then
		// SprintfBox (MB_OK, "OpenProcess Token", "Fail" );    
		return ERROR_OPEN_PROCESS_TOKEN;
	endif;                                                    
	
	// Get the Luid
	nRetVal = LookupPrivilegeValueA ( "", "SeRestorePrivilege",	&localLuid );
	if (!nRetVal) then
		//SprintfBox (MB_OK , "LookupPrivilegeValueA FAIL" , "localLuid = %d", localLuid );
		return ERROR_LOOKUP_PRIVILEGE;
	endif; 
	
	// Populate the TOKEN_PRIVILEGES struct, tp
	tp.PrivilegeCount = 1;            
	tp.Privileges.Luid.LowPart = localLuid.LowPart; 
	tp.Privileges.Luid.HighPart = localLuid.HighPart;
	tp.Privileges.Attributes = 0;	// Key difference 
	
	// Apply TOKEN PRIVILEGES
	nRetVal = AdjustTokenPrivileges ( hToken , FALSE , &tp , 0 , NULL , NULL );
	if (!nRetVal) then
		//SprintfBox (MB_OK, "AdjustTokenPrivileges FAIL","");
		return ERROR_TOKEN_NOT_DISABLED;
	endif;

	// Close Handle
	CloseHandle ( hToken );

	return 0;
end;

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   _NTProfRegHiveDelete
//
//  PURPOSE:	Deletes a specified value or key contained in a user hive.  
//
//	PARAMETERS:	nzRegType		- Registry item to delete.  Possible values are
//								  REGDB_SUBKEY or REGDB_VALUE
//				szSubKey		- Subkey (see RegDBDeleteValue InstallScript function)
//								  If nRegType is set to REGDB_SUBKEY, all subkeys and values
//								  beneath szSubKey are also deleted; szValue is ignored
//				szValue			- Name of the value to be deleted
//								  If nRegType is set to REGDB_VALUE, only the specified value
//								  is deleted 
//				nzProfileSwitch	- Numeric switch to indicate which profile to copy the file(s) to
//								  Possible values are DEFAULT_USER_PROFILE or OTHER_USER_PROFILE 
//				
//	RETURNS:	-1	- Non-NT target computer
//				(Other return values from __NTProfRegLoad and __NTProfRegUnload functions)
//	 
//	NOTE:		The use of the default user profile may not work in a roaming profile scenario
////////////////////////////////////////////////////////////////////////////// 
function NUMBER _NTProfRegHiveDelete ( nzRegType , szSubKey , szValue , nzProfileSwitch )
	STRING	sProfileListKey, sUserProfileDir , sRegProfileValue, sTempUserKeyValue;
	STRING	sTempProfileSID, sTempUserProfileDir;
	NUMBER	nType, nSize, nResult, nRetLoadKey, nRetUnloadKey, nRetDeleteKey;  
	LIST	lUserProfileSIDs;
	BOOL	bKeyPreviouslyLoaded; 
begin
	if !( SYSINFO.WINNT.bWinNT ) then
		return -1;
	endif; 
	
	// Qualify the nRegType argument
	if ( nzRegType != REGDB_SUBKEY ) && ( nzRegType != REGDB_VALUE ) then
		return ERROR_INVALID_REGDB;
	endif;
    
   	RegDBSetDefaultRoot ( HKEY_LOCAL_MACHINE );
   	sProfileListKey = "Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList";
   	
   	// Process the DEFAULT_USER_PROFILE
	if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
		if ( SYSINFO.WINNT.bWinNT4 ) then
			// Get Default User Profile (all changes to this profile will be inherited to 
			// newly created profiles on the target machine)
			sUserProfileDir = WINDIR ^ "Profiles\\Default User";
		else 
			// Get the root folder containing all profiles
			RegDBGetKeyValueEx ( sProfileListKey , "ProfilesDirectory" , nType , sRegProfileValue , nSize );
	        __NTProfResolveProfileDir ( sRegProfileValue );
			// Get Default User profile folder
			RegDBGetKeyValueEx (sProfileListKey, "DefaultUserProfile", nType, sTempUserKeyValue, nSize);
			// Append sRegProfileValue^sDefaultUserProfileDir 
			sUserProfileDir = sRegProfileValue ^ sTempUserKeyValue;
		endif;  
		
		// Load the Default User hive
		nRetLoadKey = __NTProfRegLoad ( HKEY_USERS , "DefaultUserHive" , sUserProfileDir ^ "NTUSER.DAT" );
		if ( nRetLoadKey != 0 ) then
			return nRetLoadKey;
		endif;
		
		// Delete the specified key
		RegDBSetDefaultRoot ( HKEY_USERS );
		Disable ( LOGGING ); 
		if ( nzRegType = REGDB_VALUE ) then
			RegDBDeleteValue ( "DefaultUserHive" ^ szSubKey , szValue );
		else
			RegDBDeleteKey ( "DefaultUserHive" ^ szSubKey );  
		endif;
		Enable ( LOGGING );
		
		// Unload the Default User hive
		nRetUnloadKey = __NTProfRegUnload ( HKEY_USERS , "DefaultUserHive" );
		if ( nRetUnloadKey != 0 ) then
			return nRetUnloadKey;
		endif;
    	
    	RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
    	return 0;
    endif;		

	if ( nzProfileSwitch = OTHER_USER_PROFILE ) then  
		bKeyPreviouslyLoaded = FALSE;
		// Enumerate User Profiles
		lUserProfileSIDs = ListCreate ( STRINGLIST ); 
		RegDBQueryKey ( sProfileListKey , REGDB_KEYS , lUserProfileSIDs );                               
		
		// Loop through the ProfileList                                     
		nResult = ListGetFirstString ( lUserProfileSIDs , sTempProfileSID );
		while ( nResult != END_OF_LIST )                                              
            RegDBSetDefaultRoot (HKEY_LOCAL_MACHINE);
			RegDBGetKeyValueEx ( sProfileListKey ^ sTempProfileSID, "ProfileImagePath", nType, sTempUserProfileDir, nSize);
			__NTProfResolveProfileDir ( sTempUserProfileDir );
			RegDBSetDefaultRoot ( HKEY_USERS );
			// Check to see if the SID is already loaded
			if ( RegDBKeyExist ( sTempProfileSID ) = 1 ) then		
				bKeyPreviouslyLoaded = TRUE;
			else
				// Load Hive
				nRetLoadKey = __NTProfRegLoad ( HKEY_USERS , sTempProfileSID , sTempUserProfileDir ^ "NTUSER.DAT" );
				if ( nRetLoadKey != 0 ) then
					return nRetLoadKey;
				endif;              
				bKeyPreviouslyLoaded = FALSE;
			endif;
			// Delete Registry item; don't care about whether successful or not. Exiting this function before
			// unloading the key can pose a problem for the user's profile  
			Disable (LOGGING);
			if ( nzRegType = REGDB_VALUE ) then  
				RegDBDeleteValue ( sTempProfileSID ^ szSubKey , szValue );
			else
				RegDBDeleteKey ( sTempProfileSID ^ szSubKey );
			endif;
			Enable (LOGGING);
			// Check to see if the hive was previously loaded
			if ( !bKeyPreviouslyLoaded )then
				// Unload Reg Key 
				nRetUnloadKey = __NTProfRegUnload ( HKEY_USERS , sTempProfileSID );
				if ( nRetUnloadKey != 0 ) then
					return nRetUnloadKey;
				endif; 
			endif;
   			nResult = ListGetNextString ( lUserProfileSIDs , sTempProfileSID );
   			bKeyPreviouslyLoaded = FALSE;
		endwhile;
	else
		// Invalid nzProfileSwitch
		return ERROR_INVALID_PROFILE; 
	endif;
	
    RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
    return 0;
end;

//////////////////////////////////////////////////////////////////////////////
//
//  FUNCTION:   _NTProfRegHiveSetKeyValue
//
//  PURPOSE:	Writes a specified value  contained in a user hive.  
//
//	PARAMETERS:	szKey			- Key (see RegDBSetKeyValueEx InstallScript function)
//				szName			- Value name for the value data to be set
//				nType			- Type of data to be set (see RegDBSetKeyValueEx InstallScript
//								  function for possible constants)
//				szValue			- Value to set
//				nSize			- Size, in bytes, of the data to be set (see RegDBSetKeyValueEx
//								  InstallScript function)
//				nzProfileSwitch	- Numeric switch to indicate which profile to copy the file(s) to
//								  Possible values are DEFAULT_USER_PROFILE or OTHER_USER_PROFILE 
//				bOverwrite		- Overwrite switch.  Set to TRUE to always overwrite and FALSE to
//								  not overwrite (useful when upgrading an application)
//
//	RETURNS:	-1	- Non-NT target computer
//				(Other return values from __NTProfRegLoad and __NTProfRegUnload functions)
//	 
//	NOTE:		The use of the default user profile may not work in a roaming profile scenario
////////////////////////////////////////////////////////////////////////////// 
function NUMBER _NTProfRegHiveSetKeyValue ( szKey , szName , nType , szValue , nSize , nzProfileSwitch , bOverwrite )
	STRING	sProfileListKey, sUserProfileDir , sRegProfileValue, sTempUserKeyValue;
	STRING	sTempProfileSID, sTempUserProfileDir, svValue;
	NUMBER	nvType, nvSize, nResult, nRetLoadKey, nRetUnloadKey, nRetDeleteKey;  
	LIST	lUserProfileSIDs;
	BOOL	bKeyPreviouslyLoaded; 
begin
	if !( SYSINFO.WINNT.bWinNT ) then
		return -1;
	endif; 
	
   	RegDBSetDefaultRoot ( HKEY_LOCAL_MACHINE );
   	sProfileListKey = "Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList";
   	
   	// Process the DEFAULT_USER_PROFILE
	if ( nzProfileSwitch = DEFAULT_USER_PROFILE ) then
		if ( SYSINFO.WINNT.bWinNT4 ) then
			// Get Default User Profile (all changes to this profile will be inherited to 
			// newly created profiles on the target machine)
			sUserProfileDir = WINDIR ^ "Profiles\\Default User";
		else 
			// Get the root folder containing all profiles
			RegDBGetKeyValueEx ( sProfileListKey , "ProfilesDirectory" , nvType , sRegProfileValue , nvSize );
	        __NTProfResolveProfileDir ( sRegProfileValue );
			// Get Default User profile folder
			RegDBGetKeyValueEx (sProfileListKey, "DefaultUserProfile", nvType, sTempUserKeyValue, nvSize);
			// Append sRegProfileValue^sDefaultUserProfileDir 
			sUserProfileDir = sRegProfileValue ^ sTempUserKeyValue;
		endif;  
		
		// Load the Default User hive
		nRetLoadKey = __NTProfRegLoad ( HKEY_USERS , "DefaultUserHive" , sUserProfileDir ^ "NTUSER.DAT" );
		if ( nRetLoadKey != 0 ) then
			return nRetLoadKey;
		endif;
		
		// Attempt to write to the specified key
		RegDBSetDefaultRoot ( HKEY_USERS );
		Disable (LOGGING);
		if ( bOverwrite ) then
			// Write to registry
			RegDBSetKeyValueEx ( "DefaultUserHive" ^szKey , szName , nType , szValue , nSize );		
		else
			// Do not overwrite if it exists
			if ( RegDBGetKeyValueEx ( "DefaultUserHive" ^szKey , szName , nvType , svValue , nvSize ) < 0 ) then
				// Unable to retrieve value
				RegDBSetKeyValueEx ( "DefaultUserHive" ^szKey , szName , nType , szValue , nSize );
			endif;
		endif;   
		Enable (LOGGING);
		
		// Unload the Default User hive
		nRetUnloadKey = __NTProfRegUnload ( HKEY_USERS , "DefaultUserHive" );
		if ( nRetUnloadKey != 0 ) then
			return nRetUnloadKey;
		endif;
    	
    	RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
    	return 0;
    endif;		

	if ( nzProfileSwitch = OTHER_USER_PROFILE ) then  
		bKeyPreviouslyLoaded = FALSE;
		// Enumerate User Profiles
		lUserProfileSIDs = ListCreate ( STRINGLIST ); 
		RegDBQueryKey ( sProfileListKey , REGDB_KEYS , lUserProfileSIDs );                               
		
		// Loop through the ProfileList                                     
		nResult = ListGetFirstString ( lUserProfileSIDs , sTempProfileSID );
		while ( nResult != END_OF_LIST )                                              
            RegDBSetDefaultRoot (HKEY_LOCAL_MACHINE);
			RegDBGetKeyValueEx ( sProfileListKey ^ sTempProfileSID, "ProfileImagePath", nvType, sTempUserProfileDir, nvSize);
			__NTProfResolveProfileDir ( sTempUserProfileDir );
			RegDBSetDefaultRoot ( HKEY_USERS );
			// Check to see if the SID is already loaded
			if ( RegDBKeyExist ( sTempProfileSID ) = 1 ) then		
				bKeyPreviouslyLoaded = TRUE;
			else
				// Load Hive
				nRetLoadKey = __NTProfRegLoad ( HKEY_USERS , sTempProfileSID , sTempUserProfileDir ^ "NTUSER.DAT" );
				if ( nRetLoadKey != 0 ) then
					return nRetLoadKey;
				endif;              
				bKeyPreviouslyLoaded = FALSE;
			endif;

			// Write Registry item; don't care about whether successful or not. Exiting this function before
			// unloading the key can pose a problem for the user's profile  
			Disable (LOGGING); 
			if ( bOverwrite ) then
				// Write to registry
				RegDBSetKeyValueEx ( sTempProfileSID ^ szKey , szName , nType , szValue , nSize );		
			else
				// Do not overwrite if it exists
				if ( RegDBGetKeyValueEx ( sTempProfileSID ^ szKey , szName , nvType , svValue , nvSize ) < 0 ) then
					// Unable to retrieve value
					RegDBSetKeyValueEx ( sTempProfileSID ^ szKey , szName , nType , szValue , nSize );
				endif;
			endif;   
			Enable (LOGGING);

			// Check to see if the hive was previously loaded
			if ( !bKeyPreviouslyLoaded )then
				// Unload Reg Key 
				nRetUnloadKey = __NTProfRegUnload ( HKEY_USERS , sTempProfileSID );
				if ( nRetUnloadKey != 0 ) then
					return nRetUnloadKey;
				endif; 
			endif;
   			nResult = ListGetNextString ( lUserProfileSIDs , sTempProfileSID );
   			bKeyPreviouslyLoaded = FALSE;
		endwhile;
	else
		// Invalid nzProfileSwitch
		return ERROR_INVALID_PROFILE; 
	endif;
	
    RegDBSetDefaultRoot ( HKEY_CLASSES_ROOT );
    return 0;
end;
